/*
 * out_yasapi.c
 * Copyright (C) 2015 Peter Belkner <pbelkner@snafu.de>
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston,
 * MA  02110-1301  USA
 */
#include <initguid.h>
#include <yasapi.h>

///////////////////////////////////////////////////////////////////////////////
#define YASAPI_WRITTEN
#if defined (YASAPI_WRITTEN) // {
static volatile int written;
static int video;
#endif // }
static int last_pause;

static Options options;
static Plugin plugin;
static Player player;

///////////////////////////////////////////////////////////////////////////////
#if 0 // {
BOOL WINAPI _DllMainCRTStartup(HANDLE hInst, ULONG ul_reason_for_call,
    LPVOID lpReserved)
{
  return TRUE;
}
#endif // }

static DWORD GetTime(void)
{
  return PlayerGetTime(&player,&plugin);
}

///////////////////////////////////////////////////////////////////////////////
void config(HWND hwnd)
{
  ConfigDialog(hwnd,&options);
}

void about(HWND hwnd)
{
  AboutDialog(hwnd);
}

void init()
{
  if (OptionsCreate(&options)<0)
    goto create;

  OptionsLoad(&options);

#if defined (YASAPI_DEBUG) // {
  if (options.debug)
    ConsoleOpen();
#endif // }

  DPUTS(options.debug,__FUNCTION__ "\n");
#if defined (_M_IX86_FP) // {
  DPRINTF(options.debug,"  architecture: %d\n",_M_IX86_FP);
#endif // }
  DPRINTF(options.debug,"  options: \"%s\"\n",options.path);
  DPRINTF(options.debug,"    share mode: %s\n",
      YASAPI_SHAREMODE_LABEL(options.eShareMode));
  DPRINTF(options.debug,"    startegy: %s\n",options.pull?"pull":"push");
  DPRINTF(options.debug,"    ring size: %.1f\n",options.qRingSize);
  DPRINTF(options.debug,"    ring filled: %.1f\n",options.qRingFill);
  DPRINTF(options.debug,"    shared size: %.1f\n",options.qShareSize);

  if (PluginCreate(&plugin,&options)<0) {
    DMESSAGE(options.debug,"creating plugin");
    goto plugin;
  }

  DPUTS(options.debug,"  " __FUNCTION__ " successful\n");

  return;
// cleanup:
  PluginDestroy(&plugin);
plugin:
  OptionsDestroy(&options);
create:
  return;
}

void quit()
{
  DPUTS(options.debug,__FUNCTION__ "\n");
  PluginDestroy(&plugin);
  DPUTS(options.debug,"  " __FUNCTION__ " successful\n");

#if defined (YASAPI_DEBUG) // {
  if (options.debug) {
    Sleep(5000);
    ConsoleClose();
  }
#endif // }

  OptionsDestroy(&options);
}

int open(int samplerate, int numchannels, int bitspersamp, int bufferlenms,
    int prebufferms)
{
  DPUTS(options.debug,__FUNCTION__ "\n");
  DPRINTF(options.debug,"  bufferlenms: %d\n",bufferlenms);

#if defined (YASAPI_WRITTEN) // {
  written=0;
  video=-666==bufferlenms;
  DPRINTF(options.debug,"  video: %d\n",video);
#endif // }
  last_pause=0;

  if (PlayerCreate(&player,&plugin,samplerate,numchannels,bitspersamp)<0) {
    DMESSAGE(options.debug,"creating player");
    goto player;
  }

  DPUTS(options.debug,"  " __FUNCTION__ " successful\n");

  return 1;
// cleanup:
  PlayerDestroy(&player);
player:
  return -1;
}

void close()
{
  DPUTS(options.debug,__FUNCTION__ "\n");
  PlayerDestroy(&player);
  DPUTS(options.debug,"  " __FUNCTION__ " successful\n");
}

int write(char *buf, int len)
{
#if defined (YASAPI_WRITTEN) // {
  written+=PlayerWrite(&player,buf,len);
#else // } {
  PlayerWrite(&player,buf,len);
#endif // }

  return 0;
}

int canwrite()
{
  return last_pause?0:PlayerAvailable(&player);
}

int isplaying()
{
  return 0;
}

int pause(int pause)
{
  int t=last_pause;

  DPUTS(options.debug,__FUNCTION__ "\n");

  if (last_pause!=pause) {
    PlayerPause(&player,pause);
    last_pause=pause;
  }

  return t;
}

void setvolume(int volume)
{
  DPUTS(options.debug,__FUNCTION__ "\n");

  if (-666!=volume) {
    plugin.volume=(double)volume/255;
    DPRINTF(options.debug,"  volume: %.2f\n",plugin.volume);
  }
}

void setpan(int pan)
{
}

void flush(int len)
{
  DPUTS(options.debug,__FUNCTION__ "\n");
  PlayerReset(&player);
#if defined (YASAPI_WRITTEN) // {
  written=len;
#endif // }
}
  
int getoutputtime()
{
  return GetTime();
}

int getwrittentime()
{
#if defined (YASAPI_WRITTEN) // {
  WAVEFORMATEX *pwfx;
  int t,l,ms;

  if (video&&!options.writtentime.video)
    return GetTime();
  else if (!video&&!options.writtentime.audio)
    return GetTime();
  else {
    pwfx=&player.wfxx.Format;
    t=pwfx->nSamplesPerSec*pwfx->nChannels;
    ms=written;

    if (t) {
      l=ms%t;
      ms /= t;
      ms *= 1000;
      ms += (l*1000)/t;

      ms/=(pwfx->wBitsPerSample/8);

      return ms;
    }
    else
      return 0;
  }
#else // } {
  return GetTime();
#endif // }
}

__declspec( dllexport ) Out_Module * winampGetOutModule()
{
  static Out_Module out;

  if (0==out.version) {
    out.version
      // module version (OUT_VER)
      =OUT_VER;
    out.description
      // description of module, with version string
      =PI_LABEL;
    out.id
      // module id. each input module gets its own.
      // non-nullsoft modules should be >= 65536. 
      =65;
    out.hMainWindow
      // winamp's main window (filled in by winamp)
      =NULL;
    out.hDllInstance
      // DLL instance handle (filled in by winamp)
      =NULL;
    out.Config
      // configuration dialog 
      =config;
    out.About
      // about dialog
      =about;
    out.Init
      // called when loaded
      =init;
    out.Quit
      // called when unloaded
      =quit;
    out.Open
      // returns >=0 on success, <0 on failure
      //
      // NOTENOTENOTE: bufferlenms and prebufferms are ignored in most
      // if not all output plug-ins. 
      //    ... so don't expect the max latency returned to be what you
      // asked for.
      // returns max latency in ms (0 for diskwriters, etc)
      // bufferlenms and prebufferms must be in ms. 0 to use defaults. 
      // prebufferms must be <= bufferlenms
      // pass bufferlenms==-666 to tell the output plugin that it's clock
      // is going to be used to sync video
      // out_ds turns off silence-eating when -666 is passed
      =open;
    out.Close
      // close the ol' output device.
      =close;
    out.Write
      // 0 on success. Len == bytes to write (<= 8192 always). buf is
      //   straight audio data. 
      // 1 returns not able to write (yet). Non-blocking, always.
      =write;
    out.CanWrite
      // returns number of bytes possible to write at a given time. 
      // Never will decrease unless you call Write (or Close, heh)
      =canwrite;
    out.IsPlaying
      // non0 if output is still going or if data in buffers waiting to be
      // written (i.e. closing while IsPlaying() returns 1 would truncate
      // the song
      =isplaying;
    out.Pause
      // returns previous pause state
      =pause;
    out.SetVolume
      // volume is 0-255
      =setvolume;
    out.SetPan
      // pan is -128 to 128
      =setpan;
    out.Flush
      // flushes buffers and restarts output at time t (in ms) 
      // (used for seeking)
      =flush;
    out.GetOutputTime
      // returns played time in MS
      =getoutputtime;
    out.GetWrittenTime
      // returns time written in MS (used for synching up vis stuff)
      =getwrittentime;
  }

  return &out;
}
