/*
 * yasapi_about.c
 * Copyright (C) 2015 Peter Belkner <pbelkner@snafu.de>
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston,
 * MA  02110-1301  USA
 */
#include <yasapi.h>
#include <resources.h>

///////////////////////////////////////////////////////////////////////////////
#define SLIDER_SCALE  1000
#define SLIDER_MIN    (SLIDER_SCALE*1)
#define SLIDER_MAX    (SLIDER_SCALE*8)
#define SLIDER_DELTA  (SLIDER_MAX-SLIDER_MIN)

///////////////////////////////////////////////////////////////////////////////
static double SliderGet(HWND hDlg, int nIdSlider)
{
  double pos;

  pos=SendMessage(
    GetDlgItem(hDlg,nIdSlider),
                                // _In_  HWND hWnd,
    TBM_GETPOS,                 // _In_  UINT Msg,
    0,                          // _In_  WPARAM wParam,
    0                           // _In_  LPARAM lParam
  );

  return pos/SLIDER_SCALE;
}

///////////////////////////////////////////////////////////////////////////////
static void GetRect(HWND hDlg, int nId, RECT *rc)
{
  GetWindowRect(
    GetDlgItem(hDlg,nId),       // _In_   HWND hWnd,
    rc                          // _Out_  LPRECT lpRect
  );

  ScreenToClient(
    hDlg,                       // _In_  HWND hWnd,
    (LPPOINT)&rc->left          // LPPOINT lpPoint
  );

  ScreenToClient(
    hDlg,                       // _In_  HWND hWnd,
    (LPPOINT)&rc->right         // LPPOINT lpPoint
  );
}

static void SliderSetText(HWND hDlg, int nIdStatic, int nIdSlider)
{
  double scale=SliderGet(hDlg,nIdSlider);
  char buf[128];

  sprintf(buf,"%.2f \xD7 minimum",scale);

  SetWindowTextA(
    GetDlgItem(hDlg,nIdStatic), // _In_      HWND hWnd,
    buf                         // _In_opt_  LPCTSTR lpString
  );
}

static double GetSliderScale(HWND hDlg, int nIdSlider)
{
  double q;

  // calculate the scaling factor.
  q=SendMessage(
    GetDlgItem(hDlg,nIdSlider),
                                // _In_  HWND hWnd,
    TBM_GETPOS,                 // _In_  UINT Msg,
    0,                          // _In_  WPARAM wParam,
    0                           // _In_  LPARAM lParam
  );

  q-=SLIDER_MIN;
  q/=SLIDER_DELTA;

  return q;
}

static void SliderSetWidth(HWND hDlg, int nIdSlider, int nIdParent)
{
  enum { MIN_WIDTH=27 };
  RECT rcSlider,rcParent;
  int pos,max,width;
  double q;

  // get the parent's maximum range.
  max=SendMessage(
    GetDlgItem(hDlg,nIdParent),
                                // _In_  HWND hWnd,
    TBM_GETRANGEMAX,            // _In_  UINT Msg,
    0,                          // _In_  WPARAM wParam,
    0                           // _In_  LPARAM lParam
  );

  // get the parent's position.
  pos=SendMessage(
    GetDlgItem(hDlg,nIdParent),
                                // _In_  HWND hWnd,
    TBM_GETPOS,                 // _In_  UINT Msg,
    0,                          // _In_  WPARAM wParam,
    0                           // _In_  LPARAM lParam
  );

  // calculate the scaling factor.
  q=pos-SLIDER_MIN;

  if (0.0<q)
    q/=(max-SLIDER_MIN);

  // calculate the slider's width based on the parent's width.
  GetRect(hDlg,nIdSlider,&rcSlider);
  GetRect(hDlg,nIdParent,&rcParent);

  width=MIN_WIDTH;
  width+=q*(rcParent.right-rcParent.left-MIN_WIDTH)+0.5;
  rcSlider.right=rcSlider.left+width;

  // resize the slider.
  MoveWindow(
    GetDlgItem(hDlg,nIdSlider),
                                // _In_  HWND hWnd,
    rcSlider.left,              // _In_  int X,
    rcSlider.top,               // _In_  int Y,
    width,                      // _In_  int nWidth,
    rcSlider.bottom-rcSlider.top,
                                // _In_  int nHeight,
    TRUE                        // _In_  BOOL bRepaint
  );

  // set the maximum to the parent's position.
  SendMessage(
    GetDlgItem(hDlg,nIdSlider),
                                // _In_  HWND hWnd,
    TBM_SETRANGE,               // _In_  UINT Msg,
    FALSE,                      // _In_  WPARAM wParam (redraw flag),
    MAKELONG(SLIDER_MIN,pos)    // _In_  LPARAM lParam
  );
}

static void SliderSync(HWND hDlg)
{
  SliderSetText(hDlg,IDC_STATIC_RING_SIZE,IDC_SLIDER_RING_SIZE);

  SliderSetWidth(hDlg,IDC_SLIDER_RING_FILL,IDC_SLIDER_RING_SIZE);
  SliderSetText(hDlg,IDC_STATIC_RING_FILL,IDC_SLIDER_RING_FILL);

  SliderSetWidth(hDlg,IDC_SLIDER_SHARE_SIZE,IDC_SLIDER_RING_FILL);
  SliderSetText(hDlg,IDC_STATIC_SHARE_SIZE,IDC_SLIDER_SHARE_SIZE);
}

static void SetOptions(HWND hDlg, const Options *pOptions)
{
  /////////////////////////////////////////////////////////////////////////////
  // sharemode
  CheckRadioButton(
    hDlg,                       // _In_  HWND hDlg,
    IDC_RADIO_EXCLUSIVE,        // _In_  int nIDFirstButton,
    IDC_RADIO_SHARED,           // _In_  int nIDLastButton,
    AUDCLNT_SHAREMODE_EXCLUSIVE==pOptions->eShareMode
        ?IDC_RADIO_EXCLUSIVE
        :IDC_RADIO_SHARED       // _In_  int nIDCheckButton
  );

  // push/pull
  CheckRadioButton(
    hDlg,                       // _In_  HWND hDlg,
    IDC_RADIO_PUSH,             // _In_  int nIDFirstButton,
    IDC_RADIO_PULL,             // _In_  int nIDLastButton,
    0==pOptions->pull
        ?IDC_RADIO_PUSH
        :IDC_RADIO_PULL         // _In_  int nIDCheckButton
  );

  /////////////////////////////////////////////////////////////////////////////
  // ring size
  SendMessage(
    GetDlgItem(hDlg,IDC_SLIDER_RING_SIZE),
                                // _In_  HWND hWnd,
    TBM_SETRANGE,               // _In_  UINT Msg,
    FALSE,                      // _In_  WPARAM wParam (redraw flag),
    MAKELONG(SLIDER_MIN,SLIDER_MAX)
                                // _In_  LPARAM lParam
  );

  SendMessage(
    GetDlgItem(hDlg,IDC_SLIDER_RING_SIZE),
                                // _In_  HWND hWnd,
    TBM_SETPOS,                 // _In_  UINT Msg,
    TRUE,                       // _In_  WPARAM wParam (redraw flag),
    pOptions->qRingSize*SLIDER_SCALE+0.5
                                // _In_  LPARAM lParam
  );

  SliderSetText(hDlg,IDC_STATIC_RING_SIZE,IDC_SLIDER_RING_SIZE);

  // ring fill
  SliderSetWidth(hDlg,IDC_SLIDER_RING_FILL,IDC_SLIDER_RING_SIZE);

  SendMessage(
    GetDlgItem(hDlg,IDC_SLIDER_RING_FILL),
                                // _In_  HWND hWnd,
    TBM_SETPOS,                 // _In_  UINT Msg,
    TRUE,                       // _In_  WPARAM wParam (redraw flag),
    pOptions->qRingFill*SLIDER_SCALE+0.5
                                // _In_  LPARAM lParam
  );

  SliderSetText(hDlg,IDC_STATIC_RING_FILL,IDC_SLIDER_RING_FILL);

  // share size
  SliderSetWidth(hDlg,IDC_SLIDER_SHARE_SIZE,IDC_SLIDER_RING_FILL);

  SendMessage(
    GetDlgItem(hDlg,IDC_SLIDER_SHARE_SIZE),
                                // _In_  HWND hWnd,
    TBM_SETPOS,                 // _In_  UINT Msg,
    TRUE,                       // _In_  WPARAM wParam (redraw flag),
    pOptions->qShareSize*SLIDER_SCALE+0.5
                                // _In_  LPARAM lParam
  );

  SliderSetText(hDlg,IDC_STATIC_SHARE_SIZE,IDC_SLIDER_SHARE_SIZE);

  /////////////////////////////////////////////////////////////////////////////
  // volume control
  CheckDlgButton(
    hDlg,                       // _In_  HWND hDlg,
    IDC_CHECKBOX_VOLUME,        // _In_  int nIDButton,
    pOptions->volume            // _In_  UINT uCheck
  );

  /////////////////////////////////////////////////////////////////////////////
  // writtentime audio control
  CheckDlgButton(
    hDlg,                       // _In_  HWND hDlg,
    IDC_CHECKBOX_WRITTENTIME_AUDIO,
                                // _In_  int nIDButton,
    pOptions->writtentime.audio // _In_  UINT uCheck
  );

  /////////////////////////////////////////////////////////////////////////////
  // writtentime video control
  CheckDlgButton(
    hDlg,                       // _In_  HWND hDlg,
    IDC_CHECKBOX_WRITTENTIME_VIDEO,
                                // _In_  int nIDButton,
    pOptions->writtentime.video // _In_  UINT uCheck
  );

  /////////////////////////////////////////////////////////////////////////////
  // mono2stereo control
  CheckDlgButton(
    hDlg,                       // _In_  HWND hDlg,
    IDC_CHECKBOX_MONO2STEREO,   // _In_  int nIDButton,
    pOptions->mono2stereo       // _In_  UINT uCheck
  );

  /////////////////////////////////////////////////////////////////////////////
  // debug
#if defined (YASAPI_DEBUG) // {
  CheckDlgButton(
    hDlg,                       // _In_  HWND hDlg,
    IDC_CHECKBOX_CONSOLE,       // _In_  int nIDButton,
    0==pOptions->debug          // _In_  UINT uCheck
  );
#else // } {
  EnableWindow(
    GetDlgItem(hDlg,IDC_STATIC_LABEL_CONSOLE),
                                // _In_  HWND hWnd,
    FALSE                       // _In_  BOOL bEnable
  );

  CheckDlgButton(
    hDlg,                       // _In_  HWND hDlg,
    IDC_CHECKBOX_CONSOLE,       // _In_  int nIDButton,
    TRUE                        // _In_  UINT uCheck
  );

  EnableWindow(
    GetDlgItem(hDlg,IDC_CHECKBOX_CONSOLE),
                                // _In_  HWND hWnd,
    FALSE                       // _In_  BOOL bEnable
  );
#endif // }
}

static void GetOptions(HWND hDlg, Options *pOptions)
{
#if defined (YASAPI_DEBUG) // {
  int debug;
#endif // }

  /////////////////////////////////////////////////////////////////////////////
  // sharemode
  pOptions->eShareMode=IsDlgButtonChecked(
    hDlg,                       // _In_  HWND hDlg,
    IDC_RADIO_EXCLUSIVE         // _In_  int nIDButton
  );

  pOptions->eShareMode=pOptions->eShareMode
  		?AUDCLNT_SHAREMODE_EXCLUSIVE
		:AUDCLNT_SHAREMODE_SHARED;

  /////////////////////////////////////////////////////////////////////////////
  // push / pull
  pOptions->pull=IsDlgButtonChecked(
    hDlg,                       // _In_  HWND hDlg,
    IDC_RADIO_PULL              // _In_  int nIDButton
  );

  /////////////////////////////////////////////////////////////////////////////
  pOptions->qRingSize=SliderGet(hDlg,IDC_SLIDER_RING_SIZE);
  pOptions->qRingFill=SliderGet(hDlg,IDC_SLIDER_RING_FILL);
  pOptions->qShareSize=SliderGet(hDlg,IDC_SLIDER_SHARE_SIZE);

  /////////////////////////////////////////////////////////////////////////////
  // mono2stereo control
  pOptions->mono2stereo=IsDlgButtonChecked(
    hDlg,                       // _In_  HWND hDlg,
    IDC_CHECKBOX_MONO2STEREO    // _In_  int nIDButton
  );

  /////////////////////////////////////////////////////////////////////////////
  // writtentime video control
  pOptions->writtentime.video=IsDlgButtonChecked(
    hDlg,                       // _In_  HWND hDlg,
    IDC_CHECKBOX_WRITTENTIME_VIDEO
                                // _In_  int nIDButton
  );

  /////////////////////////////////////////////////////////////////////////////
  // writtentime audio control
  pOptions->writtentime.audio=IsDlgButtonChecked(
    hDlg,                       // _In_  HWND hDlg,
    IDC_CHECKBOX_WRITTENTIME_AUDIO
                                // _In_  int nIDButton
  );

  /////////////////////////////////////////////////////////////////////////////
  // volume control
  pOptions->volume=IsDlgButtonChecked(
    hDlg,                       // _In_  HWND hDlg,
    IDC_CHECKBOX_VOLUME         // _In_  int nIDButton
  );

#if defined (YASAPI_DEBUG) // {
  /////////////////////////////////////////////////////////////////////////////
  // debug
  debug=pOptions->debug;

  pOptions->debug=FALSE==IsDlgButtonChecked(
    hDlg,                       // _In_  HWND hDlg,
    IDC_CHECKBOX_CONSOLE        // _In_  int nIDButton
  );

  OptionsSave(pOptions,debug);
#else // }
  OptionsSave(pOptions);
#endif // }
}

static INT_PTR CALLBACK ConfigProc(HWND hDlg, UINT uMsg, WPARAM wParam,
    LPARAM lParam)
{
  Options *pOptions=(Options *)GetWindowLongPtr(hDlg,GWLP_USERDATA);

  switch (uMsg) { 
  case WM_INITDIALOG:
    SetWindowLongPtr(hDlg,GWLP_USERDATA,lParam);
    SetWindowTextA(hDlg,PI_LABEL " Settings");
    SetOptions(hDlg,(Options *)lParam);
    break;
  case WM_HSCROLL: 
    SliderSync(hDlg);
    break;
  case WM_COMMAND: 
    switch (LOWORD(wParam)) { 
    case IDOK: 
      GetOptions(hDlg,pOptions);
      EndDialog(hDlg,TRUE); 
      return TRUE;
    case IDCANCEL: 
      EndDialog(hDlg,TRUE); 
      return TRUE;
    case IDC_BUTTON_DEFAULT:
      SetOptions(hDlg,GetDefaultOptions());
      return TRUE;
    default: 
      break; 
    } 

    break;
  default: 
    break; 
  } 

  return FALSE; 
}

void ConfigDialog(HWND hWndParent, Options *pOptions)
{
  HMODULE hModule=GetModuleHandleA(YASAPI_MODULE);
  LPCSTR lpTemplateName=MAKEINTRESOURCEA(IDD_CONFIG);

  DialogBoxParamA(
    hModule,              // _In_opt_  HINSTANCE hInstance,
    lpTemplateName,       // _In_      LPCTSTR lpTemplateName,
    hWndParent,           // _In_opt_  HWND hWndParent,
    ConfigProc,           // _In_opt_  DLGPROC lpDialogFunc,
    (LPARAM)pOptions      // _In_      LPARAM dwInitParam
  );
}
