/*
 * yasapi_lock.c
 * Copyright (C) 2015 Peter Belkner <pbelkner@snafu.de>
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston,
 * MA  02110-1301  USA
 */
#include <yasapi.h>

DWORD SyncCreate(Sync *pSync, SyncConfig *c)
{
#if defined (YASAPI_DEBUG) // {
  int debug=c->debug;

  pSync->debug=c->debug;
#endif // }
  pSync->state=0;
  pSync->pClient=c->pClient;

  // create mutex.
  pSync->hMutex=CreateMutex(
    NULL,           // _In_opt_  LPSECURITY_ATTRIBUTES lpMutexAttributes,
    FALSE,          // _In_      BOOL bInitialOwner,
    NULL            // _In_opt_  LPCTSTR lpName
  );

  if (NULL==pSync->hMutex) {
    DMESSAGE(debug,"creating mutex");
    goto mutex;
  }

  // end
  pSync->end.state=c->end.state;

  // read
  pSync->read.state=c->read.state;
  pSync->read.Condition=c->read.Condition;

  // read event is in non-signaled state.
  pSync->read.hEvent=CreateEvent(
    NULL,           // lpEventAttributes
    FALSE,          // bManualReset
    FALSE,          // bInitialState
    NULL            // lpName
  );

  if (NULL==pSync->read.hEvent) {
    DMESSAGE(debug,"creating read event");
    goto read;
  }

  // write
  pSync->write.state=c->write.state;
  pSync->write.Condition=c->write.Condition;

  // write event is in signaled state.
  pSync->write.hEvent=CreateEvent(
    NULL,           // lpEventAttributes
    FALSE,          // bManualReset
    TRUE,           // bInitialState
    NULL            // lpName
  );

  if (NULL==pSync->write.hEvent) {
    DMESSAGE(debug,"creating write event");
    goto write;
  }

  return 0;
// cleanup:
  CloseHandle(pSync->write.hEvent);
write:
  CloseHandle(pSync->read.hEvent);
read:
  CloseHandle(pSync->hMutex);
mutex:
  return -1;
}

void SyncDestroy(Sync *pSync)
{
  CloseHandle(pSync->write.hEvent);
  CloseHandle(pSync->read.hEvent);
  CloseHandle(pSync->hMutex);
}

///////////////////////////////////////////////////////////////////////////////
static BOOL SyncIsWait(Sync *pSync, SyncChannel *read, SyncChannel *write)
{
  if (SYNC_IS(pSync,write->state))
    return TRUE;
  else if (SYNC_IS(pSync,pSync->end.state))
    return FALSE;
  else if (NULL==read->Condition)
    return FALSE;
  else
    return 0==read->Condition(pSync->pClient);
}

////////
static void SyncWait(Sync *pSync, SyncChannel *read, SyncChannel *write)
{
  WaitForSingleObject(pSync->hMutex,INFINITE);

  while (SyncIsWait(pSync,read,write)) {
    ReleaseMutex(pSync->hMutex);
    SignalObjectAndWait(write->hEvent,read->hEvent,INFINITE,FALSE);
    WaitForSingleObject(pSync->hMutex,INFINITE);
  }

  pSync->state|=read->state;
}

static void SyncRelease(Sync *pSync, SyncChannel *read, SyncChannel *write)
{
  pSync->state&=~read->state;
  SetEvent(write->hEvent);
  ReleaseMutex(pSync->hMutex);
}

////////
void SyncWaitRead(Sync *pSync)
{
  SyncWait(pSync,&pSync->read,&pSync->write);
}

void SyncReleaseRead(Sync *pSync)
{
  SyncRelease(pSync,&pSync->read,&pSync->write);
}

////////
void SyncWaitWrite(Sync *pSync)
{
  SyncWait(pSync,&pSync->write,&pSync->read);
}

void SyncReleaseWrite(Sync *pSync)
{
  SyncRelease(pSync,&pSync->write,&pSync->read);
}
